/*************************************************************************/
/*									 */
/*	Main routine, c4.5						 */
/*	------------------						 */
/*									 */
/*************************************************************************/


#include "defns.i"
#include "types.i"
#include "gtime.i"

    /*  External data, described in extern.i  */

extern  float  BETA, GAMMA;
extern  int    ALFA;

extern ItemNo
        *NDistinctValues, /* no of distinct attribute values */
        AVC_MAX_ROW,      /* max no of rows in AVC */
        AVC_ROW;          /* actual no of rows in AVC */

extern ItemNo   MaxDistinctValues;

short		MaxAtt, MaxClass, MaxDiscrVal = 2;

ItemNo		MaxItem; 

Description	*Item;

DiscrValue	*MaxAttVal;

char		*SpecialStatus;

String	        *ClassName,
		*AttName,
		**AttValName,
		FileName = "DF";

short	        VERBOSITY = 0,
		TRIALS    = 10,
		EC_STRATEGY = 3; /* 1 strategy one, 2 strategy one+two, 
				3 strategy one+two+three */ 

Boolean	        GAINRATIO  = true,
		SUBSET     = false,
		BATCH      = true,
		UNSEENS    = false,
		PROBTHRESH = false;

ItemNo	        MINOBJS   = 2,
		WINDOW    = 0,
		INCREMENT = 0;
short
                PAVC_ROW  = 10; /* maximum percentuage of (MaxItem+1) as number
                                   of rows in AVC */

float		CF = 0.25;

Tree		*Pruned;

Boolean		AllKnown = true;

double		elapsedTree = 0,  /* Time for tree construction */ 
                elapsedSimp = 0,  /* Time for simplification, softening 
				     and unseen cases */
                elapsedIO = 0;    /* Time for I/O */

FILE	*output, *fopen();

    main(Argc, Argv)
/*  ----  */
    int Argc;
    char *Argv[];
{
    int o;
    extern char *optarg;
    extern int optind;
    Boolean FirstTime=true;
    short Best, BestTree();
    double elapsedTot, computetime();
	float memory,memoryc45;
    Attribute v;
    int AllTreeSize(), TreeSizeRaw();
	GETTIME	timea,timeb,timec; /* hold execution times */

    computetime(&timea,&timea);

    PrintHeader("decision tree generator");

    /*  Process options  */

    while ( (o = getopt(Argc, Argv, "12f:bupv:t:w:i:gsm:c:")) != EOF )
    {
	if ( FirstTime )
	{
	    printf("\n    Options:\n");
	    FirstTime = false;
	}

	switch (o)
	{
	case '1':   EC_STRATEGY = 1;
		    printf("\tEC4.5 Strategy one forced\n");
		    break;
	case '2':   EC_STRATEGY = 2;
		    printf("\tEC4.5 Strategies one+two forced\n");
		    break;
	case 'f':   FileName = optarg;
		    printf("\tFile stem <%s>\n", FileName);
		    break;
	case 'b':   BATCH = true;
		    printf("\tWindowing disabled (now the default)\n");
		    break;
	case 'u':   UNSEENS = true;
		    printf("\tTrees evaluated on unseen cases\n");
		    break;
	case 'p':   PROBTHRESH = true;
		    printf("\tProbability thresholds used\n");
		    break;
	case 'v':   VERBOSITY = atoi(optarg);
		    printf("\tVerbosity level %d\n", VERBOSITY);
		    break;
	case 't':   TRIALS = atoi(optarg);
		    printf("\tWindowing enabled with %d trials\n", TRIALS);
		    Check(TRIALS, 1, 10000);
		    BATCH = false;
		    break;
	case 'w':   WINDOW = atoi(optarg);
		    printf("\tInitial window size of %d items\n", WINDOW);
		    Check(WINDOW, 1, 1000000);
		    BATCH = false;
		    break;
	case 'i':   INCREMENT = atoi(optarg);
		    printf("\tMaximum window increment of %d items\n",
			   INCREMENT);
		    Check(INCREMENT, 1, 1000000);
		    BATCH = false;
		    break;
	case 'g':   GAINRATIO = false; 
		    printf("\tGain criterion used\n");
		    break;
	case 's':   SUBSET = true;
		    printf("\tTests on discrete attribute groups\n");
		    break;
	case 'm':   MINOBJS = atoi(optarg); /* UNCHECKED */
		    printf("\tSensible test requires 2 branches with >=%d cases\n",
			    MINOBJS);
		    Check(MINOBJS, 1, 1000000);
		    break;
	case 'c':   CF = atoi(optarg);
		    printf("\tPruning confidence level %g%%\n", CF);
		    Check(CF, Epsilon, 100);
		    CF /= 100;
		    break;
	case '?':   printf("unrecognised option\n");
		    exit(1);
	}
    }

    output = fopen("c4.5.log", "a");
    
    /*  Initialise  */

    GetNames();
    GetData(".data");
    printf("\nRead %d cases (%d attributes) from %s.data\n",
	   MaxItem+1, MaxAtt+1, FileName);
    
    switch( EC_STRATEGY )
      {
      case 3:	ALFA = 16;
	        AVC_MAX_ROW = (MaxItem+1)*Min((float)1/MaxClass,(float)PAVC_ROW/100);
		GAMMA = .75;
		BETA = 1 / MaxClass;
		break;
      case 2:   ALFA = 16; 
	        GAMMA = .75; 
		BETA = 0; 
		break;
      case 1:   ALFA = -1; 
	        GAMMA = 0; 
		BETA = 0; 
		break;
      }

    elapsedIO = computetime(&timea,&timeb);

    /*  Build decision trees  */

    if ( BATCH )
    {
	TRIALS = 1;
	OneTree();
	Best = 0;
    }
    else
    {
	Best = BestTree();
    }

    /*  Soften thresholds in best tree  */

    if ( PROBTHRESH )
    {
        computetime(&timeb,&timeb);
	printf("Softening thresholds");
	if ( ! BATCH ) printf(" for best tree from trial %d", Best);
	printf("\n");
	SoftenThresh(Pruned[Best]);
	printf("\n");
	elapsedSimp += computetime(&timeb, &timec);
	PrintTree(Pruned[Best]);
	elapsedIO += computetime(&timec,&timeb);
    }

    /*  Save best tree  */

    if ( BATCH || TRIALS == 1 )
    {
	printf("\nTree saved\n");
    }
    else
    {
	printf("\nBest tree from trial %d saved\n", Best);
    }

    computetime(&timeb,&timeb);

    SaveTree(Pruned[Best], ".tree");

    elapsedIO += computetime(&timeb,&timec);

    /*  Evaluation  */

    printf("\n\nEvaluation on training data (%d items):\n", MaxItem+1);
    Evaluate(false, Best);

    if ( UNSEENS )
    {   
        free(Item); /* free Item array */
        GetData(".test");
        printf("\nEvaluation on test data (%d items):\n", MaxItem+1);
        Evaluate(true, Best);
    }
    
    elapsedSimp += computetime(&timec, &timeb);
    elapsedTot = computetime(&timea, &timeb);

    /* Compute the amount of additional memory needed 
       by EC4.5 over C4.5 */

            /* 
	       Add:   Inf, Sup, PBestI, PBestS, PBestL, PBestH, 
                      NDistinctValues
	       Sub:   SplitGain, SplitInfo 
            */

    memory =  ( 6  * sizeof(ItemNo) + sizeof(ItemNo*) ) * (MaxAtt+1);
	
    memory -= 2 * (MaxItem+1) * sizeof(float);

           /* Add DistinctValues[] */ 
    ForEach(v,0,MaxAtt)
      if(!MaxAttVal[v])
	memory += (NDistinctValues[v] + 1) * sizeof(float);
 
    if( EC_STRATEGY > 1)
            /* Add: Position, PositionLimit */
      memory += 2 * MaxDistinctValues * sizeof(ItemNo) ; 

    if( EC_STRATEGY == 3)
           /* Add: AVC */
      memory += ( AVC_ROW * sizeof(float) + sizeof(float*) ) * (MaxClass+1); 

    /* 
       memoryc45 is the memory needed by C4.5
       
       Tested (Boolean), Gain, Info, Bar, UnknownRate 
       Freq, Freq[v], ValFreq
       ClassFreq, Slice1, Slice2
       Item, Weight, SplitGain, SplitInfo
       Size of Raw[0] and Pruned[0] trees
    */
    
    memoryc45 = 
      (5 * sizeof(float) + sizeof(short) + sizeof(Boolean) ) * (MaxAtt + 1)
      + (MaxDiscrVal+1) * ( sizeof( ItemCount *) + (MaxClass+2) * sizeof(ItemCount))  
      + (MaxClass+1)*sizeof(ItemCount)*3
      + (MaxItem + 1) * (sizeof(Description) + sizeof(AttValue)*(MaxAtt+2) + sizeof(float) * 3) 
      + AllTreeSize() * sizeof( TreeRec );

    memory = memory/memoryc45*100;
    
    fprintf(output, "\n%2d %8s %4.1f %4d %3d %8.3f %6.3f %6.3f %8.3f %5.1f%% %8d", 
	    EC_STRATEGY, FileName,  BETA, ALFA, TRIALS, elapsedTree, 
	    elapsedSimp, elapsedIO, elapsedTot, memory, TreeSizeRaw(Best));
    
    fclose(output);

    exit(0);
}


