/*************************************************************************/
/*									 */
/*	Sorting utilities						 */
/*	-----------------						 */
/*									 */
/*************************************************************************/


#include "defns.i"
#include "types.i"
#include "extern.i"

int	ALFA;  /* d/|T| <= ALFA -> use counting sort */ 
float	GAMMA; /* otherwise, |T| < Exp(4, (d/|T| - ALFA)/GAMMA) 
		  -> use counting sort */
               /* otherwise, use quicksort */

extern  ItemNo 
  *Sup,       /* Inf[a] = min index for values of attribute a */ 
  *Inf;       /* Sup[a] = max index for values of attribute a */ 

extern ItemNo *NDistinctValues;   

extern ItemCount *Weight;

ItemNo *Position,               /* Used in Counting Sort */
       *PositionLimit;

ItemNo MaxDistinctValues = 0;  /* Max number of distinct values 
                                  SetValues update it at the root */

/*************************************************************************/
/*									 */
/*	Initialise data structures for counting sort			 */
/*									 */
/*************************************************************************/

    InitialiseCountSort()
/*  ---------  */
{
  if( EC_STRATEGY == 1)
    {
      /* reset because counting sort is not used */
      MaxDistinctValues = 0;
    }
  else
    {
      /* set MaxDistinctValues to avoid huge data structures */
      MaxDistinctValues = Min( MaxDistinctValues, Max(MaxItem/2,30000) );
      
      Position = (ItemNo *) calloc(MaxDistinctValues, sizeof(ItemNo));
      PositionLimit = (ItemNo *) calloc(MaxDistinctValues, sizeof(ItemNo));
    }
}

/*************************************************************************/
/*									 */
/*  Main sort procedure: decide whether calling QuickSort or  CountSort  */
/*									 */
/*************************************************************************/

    CountingSort(Fp, Lp, Att, Exchange)
/*  ---------  */
    ItemNo Fp, Lp;
    Attribute Att;
    void (*Exchange)();
{
  ItemNo ncases,d;
  float alpha;
  unsigned int ncasesmin;
  static char FirstTime = 1;
  
  if( FirstTime )
    {
      InitialiseCountSort();
      FirstTime = 0;
    }
		
  ncases = Lp - Fp + 1;            /* Number of cases  */
  d = Sup[Att] - Inf[Att] + 1;     /* Range of indexes */
  alpha = d / ncases;
  
  if( d > MaxDistinctValues )
    {
      /* Counting sort data structures */
      /* are smaller than necessary    */

      Quicksort(Fp, Lp, Att, Exchange);
    }
  else
    {
      if( alpha <= ALFA )
	Countsort(Fp, Lp, Att); 
      else
	{
	  ncasesmin = 4 << (int)( Min((alpha - ALFA) / GAMMA, 27));
	  
	  if( ncases > ncasesmin )
	    Countsort(Fp, Lp, Att);
	  else
	    Quicksort(Fp, Lp, Att, Exchange);
	}
    }
}

/*************************************************************************/
/*									 */
/*	Sort items from Fp to Lp on attribute a.                	 */
/*	     On-the-fly Counting Sort        				 */
/*									 */
/*************************************************************************/

    Countsort(Fp, Lp, Att)
/*  ---------  */
    ItemNo Fp, Lp;
    Attribute Att;
{
  register ItemNo i,NewPosition,Aux;
  register Description Pivot, PivotTemp;
  register ItemCount PivotW, PivotWTemp;
  ItemNo Low, High;
  
  Low = Inf[Att];
  High = Sup[Att] - Low;
  
  ForEach(i, 0, High)
    Position[i] = 0;
  
  ForEach(i, Fp, Lp)
    Position[PVal(Item[i],Att) - Low]++;
  
  Aux = Fp - 1;
  ForEach(i, 0, High)
    {
      Aux += Position[i];
      Position[i]  = 
	PositionLimit[i] = Aux;
    }
  PositionLimit[High] = Lp-1;
  
  i = Fp;
  Pivot = Item[i];
  PivotW = Weight[i];
  
  while( i < Lp )
    {
      Aux = PVal(Pivot,Att) - Low;
      NewPosition  = Position[Aux];
      if( i < NewPosition )
        {
	  Position[Aux]--;
	  
	  /*   Swapping Pivot with Item[NewPosition] */
	  
	  PivotTemp = Item[NewPosition];
	  PivotWTemp = Weight[NewPosition];
	  Item[NewPosition] = Pivot;
	  Weight[NewPosition] = PivotW;
	  Pivot = PivotTemp;
	  PivotW = PivotWTemp;
	}
      else
	{
	  Item[i] = Pivot;
	  Weight[i] = PivotW;
	  i = PositionLimit[Aux] + 1;
	  Pivot = Item[i];
	  PivotW = Weight[i];
	}
    }
}

/*************************************************************************/
/*									 */
/*	Sort items from Fp to Lp on attribute a				 */
/*	Quicksort on PVal. Used after  indexing				 */
/*									 */
/*************************************************************************/

    Quicksort(Fp, Lp, Att, Exchange)
/*  ---------  */
    ItemNo Fp, Lp;
    Attribute Att;
    void (*Exchange)();
{
    register ItemNo Lower, Middle;
    register ItemNo Thresh;
    register ItemNo i;

    if ( Fp < Lp )
    {

        /* Choose the middle element as Thresh. This is better than 
	   C4.5 (using the leftmost) because, since strategy three does not
	   sort values, it may be the case that values are ordered */
 
	(*Exchange)((int)(Lp+Fp)/2, Lp);

	Thresh = PVal(Item[Lp], Att);

	/*  Isolate all items with values <= threshold  */

	Middle = Fp;

	for ( i = Fp ; i < Lp ; i++ )
	{ 
	    if ( PVal(Item[i], Att) <= Thresh )
	    { 
		if ( i != Middle ) (*Exchange)(Middle, i);
		Middle++; 
	    } 
	} 

	/*  Extract all values equal to the threshold  */

	Lower = Middle - 1;

	for ( i = Lower ; i >= Fp ; i-- )
	{
	    if ( PVal(Item[i], Att) == Thresh )
	    { 
		if ( i != Lower ) (*Exchange)(Lower, i);
		Lower--;
	    } 
	} 

	/*  Sort the lower values  */


	Quicksort(Fp, Lower, Att, Exchange);

	/*  Position the middle element  */

	(*Exchange)(Middle, Lp);

	/*  Sort the higher values  */

	Quicksort(Middle+1, Lp, Att, Exchange);
    }
}


/*************************************************************************/
/*									 */
/*	Sort items from Fp to Lp on attribute a				 */
/*	Quicksort on CPVal. Used before indexing			 */
/*									 */
/*************************************************************************/

    QuicksortCP(Fp, Lp, Att, Exchange)
/*  ---------  */
    ItemNo Fp, Lp;
    Attribute Att;
    void (*Exchange)();
{
    register ItemNo Lower, Middle;
    register float Thresh;
    register ItemNo i;

    if ( Fp < Lp )
    {
	Thresh = CPVal(Item[Lp], Att);

	/*  Isolate all items with values <= threshold  */

	Middle = Fp;

	for ( i = Fp ; i < Lp ; i++ )
	{ 
	    if ( CPVal(Item[i], Att) <= Thresh )
	    { 
		if ( i != Middle ) (*Exchange)(Middle, i);
		Middle++; 
	    } 
	} 

	/*  Extract all values equal to the threshold  */

	Lower = Middle - 1;

	for ( i = Lower ; i >= Fp ; i-- )
	{
	    if ( CPVal(Item[i], Att) == Thresh )
	    { 
		if ( i != Lower ) (*Exchange)(Lower, i);
		Lower--;
	    } 
	} 

	/*  Sort the lower values  */

	QuicksortCP(Fp, Lower, Att, Exchange);

	/*  Position the middle element  */

	(*Exchange)(Middle, Lp);

	/*  Sort the higher values  */

	QuicksortCP(Middle+1, Lp, Att, Exchange);
    }
}
